﻿//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  Cy4NET Programmlader Typ 1                                                          //
//  _________________________________                                                   //
//                                                                                      //
//  Copyright (c) 2023 Reimar Grasbon                                                   //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//  This software is subjected to the FreeBSD License. Redistribution and use in source //
//  and binary forms, with or without modification, are permitted provided that the     //
//  following conditions are met:                                                       //
//                                                                                      //
//  1. Redistributions of source code must retain the above copyright notice, this list //
//     of conditions and the following disclaimer.                                      //
//                                                                                      //
//  2. Redistributions in binary form must reproduce the above copyright notice, this   //
//     list of conditions and the following disclaimer in the documentation and/or      //
//     other materials provided with the distribution.                                  //
//                                                                                      //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY //
//  EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED           //
//  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  //
//  IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,    //
//  INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT  //
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR  //
//  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,   //
//  WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)  //
//  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE          //
//  POSSIBILITY OF SUCH DAMAGE.                                                         //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////
//                                                                                      //
//   Organisation Flashspeicher:                                                        //
//                                                                                      //
//                             +-------------------+ 0k                                 //
//                             |                   |                                    //
//                             |   Programmlader   |                                    //
//                             |                   |                                    //
//                             +-------------------+                                    //
//                             |                   |                                    //
//                             |                   |                                    //
//                             |                   |                                    //
//                             |                   |                                    //
//                             |                   |                                    //
//                             |                   |                                    //
//                             |    Applikation    |                                    //
//                             |                   |                                    //
//                             |                   |                                    //
//                             |                   |                                    //
//                             |                   |                                    //
//                             |                   |                                    //
//                             |                   |                                    //
//                             +-------------------+                                    //
//                             |      Cy4Adr       |                                    //
//                             +-------------------+ Flashende                          //
//                                                                                      //
//////////////////////////////////////////////////////////////////////////////////////////

#include "ARMChip.h"
#include "TypeDefs.h"
#include "Conf.h"
#include "Flash.h"
#include "Board.h"
#include "Cy4Message.h"
#include "Cy4TypePlate.h"

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen & Definitionen  ////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
#define PAYLOAD_CAPACITY       64 // Kapazität der PDU DAT Sektion: Fest 64 [Bytes]
#define TRANSFER_TIMEOUT      250 // Zeitfenster für Übertragung: 150 .. 250 [ms]
#define BUS_INTERBYTE_TIMEOUT   3 // Zeitüberschreit. bei Bus-Byteübertr.: 1 ..   30  [ms]
#define DEAD_PEER_TIMEOUT    3000 // Zeitgrenze bei Ausbleiben von PDUs: 1000 .. 3000 [ms]
#define PDU_DPS              0xAA // Datagramm-Einleitungssequenz: 0b10101010
#define PDU_DIB              0x30 // Datagramm-Identifikations-Bits: 0b110000
#define PDU_DTF_P               0 // Datagram Netzlast-Flag: P (Payload)
#define PDU_DTF_D               1 // Datagram Richtungs-Flag: D (Direction)

extern const UInt08 APPL_SECTION_START[8];		// Definiert im ...
extern const UInt08 CY4ADR_SECTION_START[8];	// ... Linker-Command-File

//////////////////////////////////////////////////////////////////////////////////////////
//  Typen  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
typedef struct
{   UInt08  byte[10+PAYLOAD_CAPACITY+1];
} Cy4PDU;

//////////////////////////////////////////////////////////////////////////////////////////
//  Daten  ///////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static UInt08   zero[FLASH_PAGE_SIZE]  = { 0 };
static UInt08   page[FLASH_PAGE_SIZE]  = { 0 };
static bool     progInProg             = false;      // true = Programmierung im Gange
static          Cy4Address  selfAdr    = CY4ADDRESS_INVALID;
static Cy4PDU   pdu                    = { { 0 } };
static PageNr   workingPageNr          = { 0 };
static PageNr   zeroPageNr             = { 0 };

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4NET PDU  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static Cy4MessageAccess
CY4PDU_Access(void)
{
    if (pdu.byte[1] == 0xC1 || pdu.byte[1] == 0xC2) return CY4MESSAGE_READ;
    else                                            return CY4MESSAGE_WRITE;
}

//////////////////////////////////////////////////////////////////////////////////////////
static Cy4Address
CY4PDU_Address(void)
{
    Cy4Address address = CY4ADDRESS(pdu.byte[2], pdu.byte[3]);
    return address;
}

//////////////////////////////////////////////////////////////////////////////////////////
UInt32
CY4PDU_Location(void)
{
    return (((UInt32)pdu.byte[6])<<16) + (((UInt16)pdu.byte[7])<<8) + pdu.byte[8];
}

//////////////////////////////////////////////////////////////////////////////////////////
static UInt08
CY4PDU_Count(void)
{
    return pdu.byte[9];
}

//////////////////////////////////////////////////////////////////////////////////////////
static UInt08*
CY4PDU_Data(void)
{
    return &pdu.byte[10];
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Cy4NET System ////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static bool
CY4NET_IsRequest(void)
{
    return (bool)(pdu.byte[1] & BV(PDU_DTF_D));
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
CY4NET_HasPayload(void)
{
    return (bool)(pdu.byte[1] & BV(PDU_DTF_P));
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
CY4NET_UpdateCRC(UInt08 data, UInt08* crc)
{
    *crc ^= data;
    for (UInt08 i=0; i<8; i++) *crc = (*crc >> 1) ^ ((*crc & 1) ? 0x8C : 0 );
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
CY4NET_SaveAddress(Cy4Address address)
{
	PageNr pageNr = FLASH_PageNumberOf(&CY4ADR_SECTION_START);
	UInt08 pageData[FLASH_PAGE_SIZE];
	FLASH_ReadPage(pageNr, pageData);
	UInt32 index = ((UInt32)(&CY4ADR_SECTION_START) & (FLASH_PAGE_SIZE-1));
	pageData[index+0] =  address.net;
	pageData[index+1] =  address.dev;
	pageData[index+2] = ~address.net;
	pageData[index+3] = ~address.dev;
    FLASH_WritePage(pageNr, pageData);
}

//////////////////////////////////////////////////////////////////////////////////////////
static Cy4Address
CY4NET_LoadAddress(void)
{
	const UInt08* baseAddress = (const UInt08*)(&CY4ADR_SECTION_START);
    UInt08 invData[] = { ~(baseAddress[2]), ~(baseAddress[3]) };
    Cy4Address address;
    if (baseAddress[0] == invData[0] && baseAddress[1] == invData[1])
    {   address.net = baseAddress[0];
        address.dev = baseAddress[1];
    }
    else
    {   address.net = CFG_DEFAULT_CY4_ADDRESS_NET;
        address.dev = CFG_DEFAULT_CY4_ADDRESS_DEV;
        CY4NET_SaveAddress(address);
    }
    return address;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
CY4_Setup(void)
{
    selfAdr = CY4NET_LoadAddress();
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
CY4NET_WaitForMessage(void)
{
    UInt08  i;
    UInt08  n;
    UInt08  crc;
    UInt08  byte;
	UInt32  timer = BOARD_SetTimeout(CFG_DEAD_PEER_TIMEOUT);
    do
    {   UInt08 ledPulse = 0x13;
        do
        {   if ((ledPulse&0x12) == 0x12) BOARD_InvLed();
            ledPulse++;
            BOARD_Watchdog();
            if (BOARD_Timeout(timer))
            {   // Wenn längere Zeit kein Datagramm eintrifft, Programmierung abbrechen
                progInProg = false;
            }
        } while (!BOARD_WaitForByte(&byte, 25));
        BOARD_SetLed();
        i = 0;
        n = 10 + PAYLOAD_CAPACITY;
        crc = 255;
        if (byte == PDU_DPS)
        {   do
            {   pdu.byte[i] = byte;
                if (i == 1 && (byte&0xFC) != (PDU_DIB<<2)) break;
                if (i == 9)
                {   if (byte > PAYLOAD_CAPACITY) break;
                    if (CY4NET_HasPayload()) n = 10+byte;
                    else                     n = 10;
                }
                CY4NET_UpdateCRC(byte, &crc);
                if (!BOARD_WaitForByte(&byte, BUS_INTERBYTE_TIMEOUT)) break;
                i++;
            } while (i<n);
        }
    } while (i != n                              ||  // Anzahl empfangener Bytes korrekt?
             byte != crc                         ||  // Prüfsumme stimmt?
             !CY4NET_IsRequest()                 ||  // Datagramm ist eine Anfrage?
             CY4PDU_Address().net != selfAdr.net ||  // Anfrage ist ...
             CY4PDU_Address().dev != selfAdr.dev);   //             ... an uns gerichtet?
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
CY4NET_ReplyMessage(void)
{
    // (1) Richtungswechsel, Anfrage in Antwort verwandeln
    pdu.byte[1] ^= BV(PDU_DTF_D) + BV(PDU_DTF_P);
    // (2) Quell- und Zieladresse tauschen
    pdu.byte[2] = pdu.byte[4];
    pdu.byte[3] = pdu.byte[5];
    pdu.byte[4] = selfAdr.net;
    pdu.byte[5] = selfAdr.dev;
    // (3) Prüfsumme errechnen
    UInt08  count;
    if (CY4NET_HasPayload()) count = 10 + CY4PDU_Count();
    else                     count = 10;
    UInt08 crc = 0xFF;
    for (UInt08 i=0; i<count; i++) CY4NET_UpdateCRC(pdu.byte[i], &crc);
    pdu.byte[count] = crc;
    // (4) Sobald Bus frei ist, Datagram senden
    UInt32 timer = BOARD_SetTimeout(TRANSFER_TIMEOUT);
    UInt08 i = 0;
    do
    {   UInt08  dummy;
        if( !BOARD_WaitForByte(&dummy, BUS_INTERBYTE_TIMEOUT) )
        {   for (i=0; i<=count; i++)
            {   UInt08 byte = pdu.byte[i];
                BOARD_TransmitByte(byte);
                UInt08  echo;
                if (!BOARD_WaitForByte(&echo, 1) || echo != byte) break;
            }
        }
    } while (i <= count && !BOARD_Timeout(timer));
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Programmlader  ///////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
static bool
PGMLDR_IsZeroPage(PageNr page)
{
    return page.number == zeroPageNr.number;
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PGMLDR_ReadPage(PageNr pageNr)
{
    UInt08*  buffer;

    if (PGMLDR_IsZeroPage(pageNr)) buffer = zero;
    else                           buffer = page;
    FLASH_ReadPage(pageNr, buffer);
}

//////////////////////////////////////////////////////////////////////////////////////////
static bool
PGMLDR_WritePage(PageNr pageNr)
{
    UInt08* buffer;

    if (PGMLDR_IsZeroPage(pageNr)) buffer = zero;
    else                           buffer = page;
    return FLASH_WritePage(pageNr, buffer);
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PGMLDR_WaitForNextMessage(void)
{
    CY4NET_WaitForMessage();
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PGMLDR_ProcessSystemMessage(void)
{
    if (CY4PDU_Location() < 0xFFFFF0) return;
    UInt08 locLoByte = (UInt08)CY4PDU_Location();
    UInt08 count = CY4PDU_Count();
    if (CY4PDU_Access() == CY4MESSAGE_READ)
    {   // Call typeplate
        static const UInt08  typePlate[][7] PROGMEM = {  { 'C','y','4','C','o','d','e' },
                                                         CY4TYPEPLATE(DEV_APPL_TYPE,
                                                                      DEV_APPL_VERS,
                                                                      DEV_BOARD_TYPE)
                                                      };
        UInt08 i = 0;
        UInt08 k = (locLoByte-0xF0);
        do
        {   if (k >= CARDINALITY(typePlate[1])) return;
            CY4PDU_Data()[i] = typePlate[1][k];
            k++;
            i++;
        } while (i<count);
        CY4NET_ReplyMessage();
    }
    else // CY4MSG_WRITE
    {   if (count == 1 &&
            (locLoByte == 0xF8 || locLoByte == 0xFA) &&
            CY4PDU_Data()[0] == '^')
        {   // Enter application / Enter programmloader
            if (locLoByte == 0xF8 && progInProg)
            {   PGMLDR_WritePage(workingPageNr);
                PGMLDR_WritePage(zeroPageNr);
            }
            CY4NET_ReplyMessage();
            if (locLoByte == 0xFA) BOARD_Restart();
            else                   BOARD_EnterAppl();
        }
        else if (count == 2 &&
                 locLoByte == 0xFC &&
                 (CY4PDU_Data()[0] != 0 && CY4PDU_Data()[1] != 0))
        {   // Change Cy4 address
            Cy4Address  newAdress;
            newAdress.net = CY4PDU_Data()[0];
            newAdress.dev = CY4PDU_Data()[1];
            CY4NET_ReplyMessage();
            selfAdr = newAdress;
            CY4NET_SaveAddress(newAdress);
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
static void
PGMLDR_ProcessProgrammingMessage(void)
{
	if (CY4PDU_Access() == CY4MESSAGE_WRITE)
	{	UInt32 location = CY4PDU_Location();
		UInt32 count    = CY4PDU_Count();
		if (location >= ((UInt32)(&APPL_SECTION_START)&0xFFFFFF) &&
        	location <= ((UInt32)(&CY4ADR_SECTION_START)&0xFFFFFF)-count)
    	{   if (!progInProg)
			{   progInProg = true;
				workingPageNr.number = zeroPageNr.number;
				PGMLDR_ReadPage(zeroPageNr);
				FLASH_ErasePage(zeroPageNr);
			}
        	for (UInt32 i=0; i<count; i++)
        	{   PageNr accPageNr = FLASH_PageNumberOf((const void*)(FLASH_BASE+location));
        		if (accPageNr.number != workingPageNr.number)
				{   if (!PGMLDR_IsZeroPage(workingPageNr))
					{   if (!PGMLDR_WritePage(workingPageNr)) return;  // Schreibfehler
					}
					if (!PGMLDR_IsZeroPage(accPageNr)) PGMLDR_ReadPage(accPageNr);
					workingPageNr = accPageNr;
				}
				UInt32 index = (location & (FLASH_PAGE_SIZE-1));
				if (PGMLDR_IsZeroPage(workingPageNr)) zero[index] = CY4PDU_Data()[i];
				else                                  page[index] = CY4PDU_Data()[i];
				location++;
			}
        	CY4NET_ReplyMessage();
        }
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
//  Funktionen  //////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////
void
main(void)
{
	zeroPageNr = FLASH_PageNumberOf(&APPL_SECTION_START);
    BOARD_Setup();
    CY4_Setup();
    while (true)
    {   PGMLDR_WaitForNextMessage();
        PGMLDR_ProcessSystemMessage();
        PGMLDR_ProcessProgrammingMessage();
    }
}

//////////////////////////////////////////////////////////////////////////////////////////
